import datetime
import time

class LoggerException(Exception):
    pass


class Logger:
    def __init__(self, do_print=True):
        self.messages = []
        self.do_print = do_print
        self.levels = ["ok", "fail", "warn", "info", "debug"]

    def pretty_print(self, message):
        level, timestamp, content = message
        timecode = datetime.datetime.fromtimestamp(timestamp).strftime("%Y-%m-%d %H-%M-%S")
        print(f"[{level}][{timecode}] {content}")

    def get_messages(self):
        return self.messages

    def get_messages_on_level(self, wanted_level):
        return [(level, timestamp, content) for level, timestamp, content in self.messages if level == wanted_level]

    def log(self, level: str, content: str):
        if level not in self.levels:
            level_string = ", ".join(self.levels)
            raise LoggerException(f"Unknown Log level {level}, should be one of {level_string}")

        timestamp = time.time()
        message = (level, timestamp, content)
        self.messages.append(message)

        if self.do_print:
            self.pretty_print(message)

    def info(self, message: str):
        self.log("info", message)

    def warn(self, message: str):
        self.log("warn", message)

    def fail(self, message: str):
        self.log("fail", message)

    def success(self, message: str):
        self.log("ok", message)

    def exception(self, exception: Exception):
        self.fail(f"{exception.__class__}: {exception}")

    def log_list(self, messages):
        self.messages += messages

    def merge(self, logger):
        self.log_list(logger.messages)

    def analyze_log(self):
        for level in ["ok", "fail", "warn"]:
            self.info(f"{level} - {len(self.get_messages_on_level(level))}")

    def get_log_contents(self):
        return "\n".join(self.get_messages())



import copy
from sqlalchemy import MetaData, select, delete

from jhdata.filehelper import *
from jhdata.data_tools.database import create_table_if_not_exists, insert_missing, insert_one
from jhdata.dates import current_timestamp


class RequestQueue:
    def __init__(self, table_name: str, schema: str = "request_queues", filehelper: FileHelper = None):
        self.table_name = table_name
        self.schema = schema
        self.filehelper = filehelper if filehelper is not None else Files
        self.meta = MetaData(self.filehelper.get_engine())
        self.sleep_duration = 3
        self.request_queue_schema = self.filehelper.read_schema("request_queue")

        # make sure the table exists
        self.get_table()

    def get_table(self):
        return create_table_if_not_exists(self.table_name, self.request_queue_schema, self.schema, self.filehelper)

    def fetch_one(self, use_secrets=True):
        table = self.get_table()
        statement = select(table).order_by(table.c.priority.desc(), table.c.requested_at.desc())
        row = self.filehelper.sql(statement).first()

        if row is None:
            return None, None
        else:
            # Convert to dict to allow assignment later
            row = {key: row[key] for key in row.keys()}
            # Copy row to be able to mark it as done later
            row_raw = copy.deepcopy(row)
            # Substitute secret values and load JSON
            if use_secrets:
                row["metadata"] = self.filehelper.secrets.substitute(row["metadata"])

            row["metadata"] = json.loads(row["metadata"])
            return row, row_raw

    def mark_as_done(self, row: dict):
        table = self.get_table()
        statement = delete(table)

        for key in row.keys():
            statement = statement.where(table.c[key] == row[key])

        return self.filehelper.sql(statement)

    def insert_many(self, items):
        DF_THRESHOLD = 10

        print(items[:3])
        items_parsed = [{
            "table_name": item["table_name"],
            "metadata": json.dumps(item["metadata"]),
            "priority": item["priority"],
            "requested_at": current_timestamp()
        } for item in items]

        if len(items_parsed) < DF_THRESHOLD:
            for item in items_parsed:
                insert_one(
                    item,
                    self.table_name,
                    schema_name=self.schema,
                    table_schema=self.request_queue_schema,
                    skip_if_pk_match=True
                )
        else:
            self.filehelper.logger.info(f"Queuing in bulk {len(items_parsed)} items")
            item_df = pd.DataFrame(items_parsed)
            insert_missing(item_df, self.table_name, schema_name=self.schema, table_schema=self.request_queue_schema)

    def insert_one(self, table_name: str, metadata: dict, priority: int = 0):
        return self.insert_many([
            {
                "table_name": table_name,
                "metadata": metadata,
                "priority": priority,
                "requested_at": current_timestamp()
            }
        ])



from jhdata.filehelper import FileHelper
from jhdata.logger import Logger
from jhdata.loads.LoadStep import LoadStep
from jhdata.dates import current_timestamp


class Load:
    def __init__(self, source_name: str, steps: list[LoadStep] = None, input=None, filehelper=None, logger=None):
        self.steps = steps if steps is not None else []
        self.files = filehelper if filehelper is not None else FileHelper()
        self.logger = logger if logger is not None else Logger()
        self.input = input
        self.source_name = source_name

    def run(self):
        result_dict = {}
        current_data = self.input

        for step in self.steps:
            step.load = self
            data_out, metadata = step.run(current_data)

            current_data = data_out
            result_dict[step.step_name] = metadata

        self.finalize()

        return current_data, result_dict

    def finalize(self):
        self.logger.analyze_log()
        log_content = self.logger.get_log_contents()
        log_path = f"logs/{self.source_name}/log_{current_timestamp()}"
        self.files.write_text(log_path, log_content)



import pandas as pd
from sqlalchemy import inspect, exists
from sqlalchemy.schema import CreateSchema
from sqlalchemy import Table, MetaData, select, insert, delete

from jhdata.filehelper import Files
from jhdata.data_tools.dataframes import *


def create_schema_if_not_exists(schema_name: str, filehelper=None):
    filehelper = filehelper if filehelper is not None else Files
    engine = filehelper.get_engine()
    with engine.connect() as conn:
        if schema_name not in conn.dialect.get_schema_names(conn):
            conn.execute(CreateSchema(schema_name))


def create_table_if_not_exists(table_name: str, table_schema = None, schema_name: str = None, filehelper = None):
    filehelper = filehelper if filehelper is not None else Files
    table_schema = table_schema if table_schema is not None else filehelper.read_table_schema(table_name, schema_name)
    engine = filehelper.get_engine()

    inspector = inspect(engine)

    if schema_name is not None:
        create_schema_if_not_exists(schema_name, filehelper)

    if not inspector.has_table(table_name, schema=schema_name):
        filehelper.write_sql(table_name, table_schema.make_df(), schema=schema_name)

    return Table(table_name, MetaData(engine), autoload_with=engine, schema=schema_name)


def table_exists(table_name: str, schema_name: str = None, filehelper = None):
    filehelper = filehelper if filehelper is not None else Files
    engine = filehelper.get_engine()

    inspector = inspect(engine)
    return inspector.has_table(table_name, schema=schema_name)


def upsert(update_df: pd.DataFrame, table_name: str, table_schema = None, schema_name: str = None, filehelper = None):
    filehelper = filehelper if filehelper is not None else Files
    table_schema = table_schema if table_schema is not None else filehelper.read_table_schema(table_name, schema_name)

    create_table_if_not_exists(table_name, table_schema, schema_name, filehelper)

    current_df = filehelper.read_sql_table(table_name, schema=schema_name)
    updated_df = upsert_dataframes(current_df, update_df, table_schema)

    validated_df = table_schema.validate(updated_df)

    return filehelper.write_sql(table_name, validated_df, schema=schema_name, table_schema=table_schema)


def insert_missing(update_df: pd.DataFrame, table_name: str, table_schema = None, schema_name: str = None, filehelper = None):
    filehelper = filehelper if filehelper is not None else Files
    table_schema = table_schema if table_schema is not None else filehelper.read_table_schema(table_name, schema_name)

    validated_df = table_schema.validate(update_df)

    create_table_if_not_exists(table_name, table_schema, schema_name, filehelper)

    current_df = filehelper.read_sql_table(table_name, schema=schema_name)
    insert_df = append_missing_values(current_df, validated_df, table_schema)

    return filehelper.write_sql(table_name, insert_df, schema=schema_name, table_schema=table_schema)


def insert_one(item: dict, table_name: str, table_schema = None, skip_if_pk_match = False, schema_name: str = None, filehelper = None):
    filehelper = filehelper if filehelper is not None else Files
    table_schema = table_schema if table_schema is not None else filehelper.read_table_schema(table_name, schema_name)

    table = create_table_if_not_exists(table_name, table_schema, schema_name, filehelper)

    if skip_if_pk_match:
        statement = exists(table)
        for key in table_schema.primary_keys():
            statement = statement.where(table.c[key] == item[key])

        statement = select(statement)
        item_exists = filehelper.sql(statement).fetchone()[0]
        if item_exists:
            return

    statement = insert(table).values(**item)

    return filehelper.sql(statement)



import io
import os
import re

import requests
import boto3
import s3fs
import sqlalchemy

from jhdata.logger import *
from jhdata.data_tools.schemas import *
from jhdata.secrets import SecretInterfaceMemory

"""
Useful stuff:
JSON schemas: https://specs.frictionlessdata.io/table-schema/#types-and-formats
"""


class FileHelper:
    def __init__(self, bucket_name: str = None, logger: Logger = None, environment: str = None, engine = None, secrets = None):
        self.environment = environment if environment else os.getenv("ENVIRONMENT", "dev")
        self.bucket_name = bucket_name if bucket_name else self.environment
        self.logger = logger if logger else Logger()
        self.engine = engine
        self.fs = None
        self.boto = None
        self.secrets = secrets if secrets is not None else SecretInterfaceMemory()

    def set_engine(self, engine):
        self.engine = engine

    def get_engine(self):
        if isinstance(self.engine, sqlalchemy.engine.Engine):
            return self.engine
        else:
            print(self.engine)
            raise Exception("No database engine defined, use connect_db() first")

    def connect_db(self, connection_string: str):
        self.set_engine(sqlalchemy.create_engine(connection_string))
        return self.get_engine()

    def connect_s3(self,
                   endpoint_url: str,
                   aws_access_key_id: str,
                   aws_secret_access_key: str,
                   aws_session_token: str = None):
        s3fs_config = {
            "anon": False,
            "key": aws_access_key_id,
            "secret": aws_secret_access_key,
            "token": aws_session_token,
            "use_ssl": False,
            "client_kwargs": {
                "endpoint_url": endpoint_url,
            }
        }

        boto3_config = {
            "service_name": "s3",
            "endpoint_url": endpoint_url,
            "aws_access_key_id": aws_access_key_id,
            "aws_secret_access_key": aws_secret_access_key,
            "aws_session_token": aws_session_token,
            "verify": True
        }

        print("s3fs_config:", s3fs_config)
        print("boto3_config:", boto3_config)

        self.fs = s3fs.S3FileSystem(**s3fs_config)
        self.boto = boto3.resource(**boto3_config)

        return self.fs, self.boto

    @property
    def bucket(self):
        if self.boto is None:
            raise Exception("No connection established, call connect_s3() first")

        return self.boto.Bucket(self.bucket_name)

    def sql(self, statement, **kwargs):
        with self.get_engine().connect() as connection:
            with connection.begin():
                return connection.execute(statement, **kwargs)

    # Get path for s3fs
    def fspath(self, path: str):
        return f"{self.bucket_name}/{path}"

    def metadata(self, path: str):
        return self.fs.metadata(self.fspath(path))

    # Last modified as datetime
    def last_modified(self, path: str) -> datetime.datetime:
        return self.fs.modified(self.fspath(path))

    # Check if path is a directory
    def is_directory(self, path: str) -> bool:
        return self.fs.isdir(self.fspath(path))

    # Check if a path exists
    def path_exists(self, path: str) -> bool:
        return self.fs.exists(self.fspath(path))

    # Check if a path exists and is a directory
    def directory_exists(self, path: str) -> bool:
        return self.path_exists(path) and self.is_directory(path)

    # Check if a path exists and is a file
    def file_exists(self, path: str) -> bool:
        return self.path_exists(path) and not self.is_directory(path)

    def find(self, path: str, **kwargs):
        return [obj.key for obj in self.bucket.objects.filter(Prefix=path, **kwargs)]

    def ls(self, path: str, **kwargs):
        return self.fs.ls(path=self.fspath(path), **kwargs)

    def find_regex(self, path: str, regex: str, **kwargs):
        files = self.find(path, **kwargs)
        results = []

        for file in files:
            filename = file.split("/")[-1]
            match = re.search(regex, filename)
            if match is not None:
                results.append((file, match.groupdict()))

        return results

    # Generate an URL to access a file
    def url(self, path: str, **kwargs):
        return self.fs.url(path=self.fspath(path), **kwargs)

    # Move file
    def move(self, path_from: str, path_to: str, recursive: bool = False):
        self.fs.move(self.fspath(path_from), self.fspath(path_to), recursive=recursive)

    # Copy file
    def copy(self, path_from: str, path_to: str, recursive: bool = False):
        self.fs.copy(self.fspath(path_from), self.fspath(path_to), recursive=recursive)

    # Delete file
    def delete(self, path: str, recursive: bool = False):
        self.fs.rm(self.fspath(path))

    # Get a readable file object
    def get_readable(self, path: str):
        return self.fs.open(self.fspath(path))

    # Get a readable file object
    def get_readable_bytes(self, path: str):
        return self.fs.open(self.fspath(path), "rb")

    # Get a writeable file object
    def get_writeable(self, path: str):
        return self.fs.open(self.fspath(path), "w")

    # Get a writeable file object
    def get_writeable_bytes(self, path: str):
        return self.fs.open(self.fspath(path), "wb")

    # Download a file into a ByteIO object
    def read_byteio(self, path: str):
        output = io.BytesIO()
        self.boto.Bucket(self.bucket_name).download_fileobj(Key=path, Fileobj=output)
        return output

    # Download file content as bytes
    def read_bytes(self, path: str):
        output = self.read_byteio(path)
        return output.getvalue()

    # Upload bytes into a file
    def write_bytes(self, path: str, data: bytes):
        self.logger.info(f"[{self.bucket_name}] Uploading bytes to {path}")
        result = self.boto.Bucket(self.bucket_name).put_object(Key=path, Body=data)
        return result

    # Download a file from an URL to bucket using HTTP
    def download_file(self, url: str, to_path: str, **kwargs):
        self.logger.info(f"[{self.bucket_name}] Reading file bytes from {url}")
        response = requests.get(url=url, **kwargs)
        self.write_bytes(to_path, response.content)

    # Download file content as text
    def read_text(self, path: str, encoding: str = "utf8"):
        byte_content = self.read_bytes(path)
        return byte_content.decode(encoding)

    # Upload text into a file
    def write_text(self, path: str, text: str, encoding: str = "utf8"):
        self.logger.info(f"[{self.bucket_name}] Uploading text to {path}")
        result = self.boto.Bucket(self.bucket_name).put_object(Key=path, Body=text.encode(encoding))
        return result

    def read_json_content(self, path: str, encoding: str = "utf8"):
        byte_content = self.read_bytes(path)
        text_content = byte_content.decode(encoding)
        return json.loads(text_content)

    # Read a JSON file into a Pandas DataFrame
    def read_json(self, path: str, schema: Schema = None, **kwargs):
        if isinstance(schema, Schema):
            return pd.read_json(self.get_readable(path), dtype=schema.dtype_dict, **kwargs)
        elif schema is None:
            return pd.read_json(self.get_readable(path), **kwargs)
        else:
            raise SchemaException("Expected an instance of Schema, got", schema)

    def json_normalize(self, data: list, **kwargs):
        return pd.json_normalize(data, **kwargs)

    # Write to a JSON file
    def write_json(self, path: str, df: pd.DataFrame, **kwargs):
        return df.to_json(self.get_writeable(path), **kwargs)

    # Read a Parquet file into a Pandas DataFrame
    def read_parquet(self, path: str, table_schema: Schema = None, **kwargs):
        return pd.read_parquet(self.get_readable(path), **kwargs)

    # Write to a parquet file
    def write_parquet(self, path: str, df: pd.DataFrame, **kwargs):
        return df.to_parquet(self.get_writeable_bytes(path), **kwargs)

    # Read a SQL table into a DataFrame
    def read_sql(self, table_name: str, table_schema: Schema = None, **kwargs):
        engine = self.get_engine()
        return pd.read_sql(table_name, engine, **kwargs)

    # Read a SQL table into a DataFrame
    def read_sql_table(self, table_name: str, table_schema: Schema = None, **kwargs):
        with self.get_engine().begin() as conn:
            return pd.read_sql_table(table_name, conn, **kwargs)

    # Write a DataFrame into a SQL table
    def write_sql(self, table_name: str, df: pd.DataFrame, table_schema: Schema = None, if_exists="replace", index=False, **kwargs):
        engine = self.get_engine()
        result = df.to_sql(table_name, engine, if_exists=if_exists, index=index, **kwargs)
        self.logger.success(f"Wrote to table {table_name} - result: {result}")

    # Read schema from disk
    def read_schema(self, path: str):
        schema_dict = json.loads(self.read_text(f"schemas/{path}.json"))
        return Schema(schema_dict)

    # Read schema from default location for a SQL table
    def read_table_schema(self, table_name: str, schema_name: str = None):
        if schema_name is None:
            return self.read_schema(table_name)
        else:
            return self.read_schema(f"{schema_name}/{table_name}")

    # Write schema to disk
    def write_schema(self, path: str, schema: Schema):
        schema_string = schema.to_json()
        self.write_text(f"schemas/{path}.json", schema_string)


Files = FileHelper()



"""
Reader classes are meant to read a delivery of data, which can contain multiple tables

They take an arbitrary input and return a dict in a format like {"table1": dataframe1, "table2": dataframe2}
"""

import os
import pandas as pd


# Takes a dict of {table: dataframe} directly in Memory and just returns it
class Reader:
    def read(self, load_step, input: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
        return input


# Reads parquet files from a directory
class JSONReader(Reader):
    def __init__(self, filehelper=None, table_regex=r"(?P<table_name>.*)_(?P<timestamp>[0-9]{8}_[0-9]{6})\.json", directory=None, **json_args):
        self.filehelper = filehelper
        self.directory = directory
        self.json_args = json_args
        self.table_regex = table_regex

    def read(self, load_step, input) -> dict[str, pd.DataFrame]:
        fh = self.filehelper if self.filehelper else load_step.filehelper
        path = self.directory if self.directory is not None else f"raw/{load_step.load.source_name}/staging"

        load_step.logger.info(f"Looking for valid files in {path}")
        valid_files = fh.find_regex(path, self.table_regex)
        print(valid_files)

        table_dict = {}
        for file_path, groups in valid_files:
            table_dict[groups["table_name"]] = fh.read_json(file_path, **self.json_args)
            load_step.logger.info(f"Read JSON file at {file_path}")

        return table_dict


class SQLReader(Reader):
    def __init__(self, filehelper=None, table=None, **sql_args):
        self.filehelper = filehelper
        self.table = table




"""
Writer classes are meant to write a delivery of data to a target, which can contain multiple tables

They take a dict in a format like {"table1": dataframe1, "table2": dataframe2}
"""

from jhdata.loads.helpers import *


# Base Writer class
class Writer:
    def write(self, load_step, input: dict[str, pd.DataFrame]):
        load_step.logger.fail("The Writer class should not be used directly, but inherited from. "
                               "You can use NoopWriter instead")


# Writer class that does nothing on write
class NoopWriter(Writer):
    def write(self, load_step, input: dict[str, pd.DataFrame]):
        load_step.logger.info("Write step skipped")


# Reads parquet files from a directory
class ParquetWriter(Writer):
    def __init__(self, filehelper=None, directory=None, **parquet_args):
        self.filehelper = filehelper
        self.directory = directory
        self.parquet_args = parquet_args

    def write(self, load_step, input: dict[str, pd.DataFrame]) -> None:
        fh = self.filehelper if self.filehelper else load_step.filehelper

        for table_name, df in input.items():
            path = f"{self.directory}/{table_name}" if self.directory is not None else f"raw/{load_step.load.source_name}/staging/{table_name}.parquet"
            fh.write_parquet(path, df, **self.parquet_args)
            size = get_pd_statistics(df)["size"]
            load_step.logger.info(f"Wrote file to {path} with {size}")



import pandas as pd
import numpy as np


from jhdata.data_tools.schemas import *
from jhdata.logger import Logger
from jhdata.dates import current_datetime


def add_load_timestamp(df: pd.DataFrame):
    if "load_timestamp" not in df.columns:
        df["load_timestamp"] = current_datetime()

    return df


def upsert_dataframes(target_df: pd.DataFrame, update_df: pd.DataFrame, schema: Schema, logger: Logger = None):
    logger = logger if logger is not None else Logger()

    pks = schema.primary_keys
    update_df = add_load_timestamp(update_df)

    logger.info(f"Upserting dataframe with {update_df.shape[0]} rows...")

    update_df = update_df.set_index(pks, drop=False)
    target_df = target_df.set_index(pks, drop=False)

    result_df = pd.concat([target_df, update_df[~update_df.index.isin(target_df.index)]])
    result_df.update(update_df)

    logger.info(f"{result_df.shape[0] - target_df.shape[0]} were added to the table. {result_df.shape[0]} rows total")

    return result_df


def apply_delete_columns(target_df: pd.DataFrame, update_df: pd.DataFrame, schema: Schema):
    delete_columns = schema.delete_columns()
    print("delete_columns:", delete_columns)
    delete_values = set([tuple(row) for row in update_df[delete_columns].values])
    print(delete_values)
    matched_mask = np.array([tuple(row) in delete_values for row in target_df[delete_columns].values])
    result_df = target_df[~matched_mask]
    print("Dropping values...")
    print(result_df.head())
    result_df = pd.concat([result_df, update_df], ignore_index=True)
    print("Concatenating...")
    return result_df


def append_missing_values(target_df: pd.DataFrame, update_df: pd.DataFrame, schema: Schema):
    pks = schema.primary_keys

    update_df = add_load_timestamp(update_df)
    print(update_df.head())
    update_df = update_df.set_index(pks, drop=False)
    target_df = target_df.set_index(pks, drop=False)

    print("Appending...")
    result_df = pd.concat([target_df, update_df[~update_df.index.isin(target_df.index)]])

    return result_df


def merge_dataframes(target_df: pd.DataFrame, update_df: pd.DataFrame, schema: Schema):
    if len(schema.primary_keys()) > 0:
        return upsert_dataframes(target_df, update_df, schema)
    else:
        return apply_delete_columns(target_df, update_df, schema)



import datetime


class DateFormats:
    date_path = "%Y/%m/%d"
    date = "%Y-%m-%d"
    timestamp = "%Y%m%d_%H%M%S"
    datetime = "%Y-%m-%d %H:%M:%S"


def archive_date_today():
    return datetime.datetime.today().strftime(DateFormats.date_path)


def current_date():
    return datetime.datetime.today().strftime(DateFormats.date)


def current_timestamp():
    return datetime.datetime.now().strftime(DateFormats.timestamp)


def current_datetime():
    return datetime.datetime.now().strftime(DateFormats.datetime)


def backdated_timestamp(**kwargs):
    return (datetime.datetime.now() - datetime.timedelta(**kwargs)).strftime(DateFormats.timestamp)


def zero_timestamp():
    return datetime.datetime.fromtimestamp(0).strftime(DateFormats.timestamp)


def timestamp_to_epoch(timestamp):
    return int(datetime.datetime.strptime(timestamp, DateFormats.timestamp).timestamp())



import json
import pandas as pd
from pandera.io import from_json as pandera_from_json

"""
Schema formats:

Field: dict
{
    "name": "firstname",
    "dtype": "string",
    "primary": False,
    "nullable": True
}

Schema format: dict

{
    version: "1",
    fields: Field[]
}

Some notes: SQLAlchemy only references precision and length when creating tables: https://docs.sqlalchemy.org/en/14/core/tutorial.html

"""


class SchemaException(Exception):
    pass


class Schema:
    def __init__(self, schema_dict: dict, validate=True):
        if "pandera" not in schema_dict:
            print(json.dumps(schema_dict, indent=4))
            raise SchemaException("No pandera schema in schema dict")

        self.pandera = pandera_from_json(json.dumps(schema_dict["pandera"]))
        self.primary_keys = schema_dict.get("primary_keys", [])
        self.delete_columns = schema_dict.get("delete_columns", [])

        if validate:
            self.validate_self()

    def to_json(self):
        schema_dict = {
            "primary_keys": self.primary_keys,
            "delete_columns": self.delete_columns,
            "pandera": self.pandera.to_json()
        }

        return json.dumps(schema_dict, indent=4)

    def __str__(self):
        return self.to_json()

    def make_df(self):
        return pd.DataFrame({key: pd.Series(dtype=dtype) for key, dtype in self.dtype_dict.items()})

    @property
    def field_names(self):
        return [col for col in self.pandera.columns]

    @property
    def dtype_dict(self):
        return {key: str(dtype) for key, dtype in self.pandera.dtypes.items()}

    def validate_self(self):
        has_deletes = len(self.delete_columns) > 0
        has_pks = len(self.primary_keys) > 0

        if has_deletes and has_pks:
            raise SchemaException("Schema can only have one of delete columns or primary keys")

    def validate(self, df: pd.DataFrame):
        return self.pandera.validate(df)



from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("jhdata/version.txt", "r", encoding="utf-8") as fh:
    VERSION = fh.read()

NAME = "jhdata"
DESCRIPTION = "Abstractions around cloud file interfaces (WIP)"
URL = "https://github.com/jothapunkt/jhdata"
EMAIL = "jakob-hoefner@web.de"
AUTHOR = "Jothapunkt"
REQUIRED = [
    "requests",
    "boto3",
    "s3fs",
    "pandas",
    "pyarrow",
    "sqlalchemy",
    "psycopg2",
    "pandera[io]",
    "aioitertools",
    "pyrsistent"
]

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=EMAIL,
    install_requires=REQUIRED,
    url=URL,
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.8",
    include_package_data=True
)



import pandas as pd

from jhdata.logger import Logger
from jhdata.loads.Reader import Reader
from jhdata.loads.Writer import Writer
from jhdata.filehelper import FileHelper
from jhdata.loads.helpers import get_pd_statistics

"""
One step of a load - meant to be inherited from
By default receives a DataFrame in Memory and writes it as a Parquet file only if write_to_disk is set to True
"""


class LoadStep:
    def __init__(self, step_name, reader: Reader = Reader(), writer: Writer = Writer(),
                 transformation=lambda data: data, filehelper: FileHelper = None):
        self.load = None
        self.step_name = step_name
        self.filehelper = filehelper if filehelper is not None else FileHelper()
        self.logger = Logger()

        self.reader = reader
        self.transformation = transformation
        self.writer = writer

    """
    Call the other relevant methods in order:
    - load
    - process
    - write
    
    Then return the data and some metadata if applicable
    """

    def run(self, input):
        data_in = self.reader.read(self, input)
        data_out = self.transformation(data_in)

        self.writer.write(self, data_out)

        # Can be expanded in the future by using a more general metadata function
        if isinstance(data_out, pd.DataFrame):
            return data_out, get_pd_statistics(data_out)
        else:
            return data_out, None
